//////////////////////////////////////////////////////////////////
//
//    Fries - Feature Retriever for Intensional Encoding of Sentences
//
//    Copyright (C) 2006   TALP Research Center
//                         Universitat Politecnica de Catalunya
//
//    This file is part of the Fries library
//
//    The Fries library is free software; you can redistribute it 
//    and/or modify it under the terms of the GNU General Public
//    License as published by the Free Software Foundation; either
//    version 2 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    General Public License for more details.
//
//    You should have received a copy of the GNU General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 51 Franklin St, 5th Floor, Boston, MA 02110-1301 USA
//
//    contact: Lluis Padro (padro@lsi.upc.es)
//             TALP Research Center
//             despatx Omega.S112 - Campus Nord UPC
//             08034 Barcelona.  SPAIN
//
////////////////////////////////////////////////////////////////

//===========================================================
//=                                                         =
//=   This module is based on FEX project developed         =
//=   at U.Illinois, and distributed under LGPL             =
//=   with the permission of its authors                    =
//=                                                         =
//=   For info about original version, visit                =
//=      http://l2r.cs.uiuc.edu/~cogcomp/                   =
//=   or contact                                            =
//=      University of Illinois at Urbana/Champaign         =
//=      Department of Computer Science                     =
//=      Dr. Dan Roth - Cognitive Computation Group         =
//=                                                         =
//===========================================================

#include <fstream>

#include "fries/fex.h"
#include "fries/readFeatures.h"

using namespace std;

const char TARGET_STRING[] = "*";
const char TARGET_CHAR = '*';
const char FILLER_STRING[] = "_";
const char FILLER_CHAR = '_';


/////////////////////////////////////////////////////////
/// Constructor:
/// Create Fex with given rules file, and load given 
/// lexicon, if any.
////////////////////////////////////////////////////////

fex::fex(const std::string &fname, const std::string &flex) {

  vector<RGF> *scr = DoParse(fname.c_str());
  if (scr) {
    rules = *scr;
    delete scr;
    lexicon_usable = false;
    lexicon.clear();
    sum=0;
  }
  else {
    cerr << "Parsing error in script file "<<fname<<endl;
    exit(1);
  }

  if (flex!="") load_lexicon(flex);
}


////////////////////////////////////////////////////////
/// encode given sentence in feature names, add
/// result to given vector. 
////////////////////////////////////////////////////////

void fex::encode_name(const sentence &s, std::vector<std::set<std::string> > &result, bool build) {
  vector<RGF>::iterator pRel;
  set<string>::iterator pFeat;
  set<string> features;

  // copy sentence (list<word>) to vector<word>, needed by FEX classes
  vector<word> sent;
  sent.reserve(s.size());
  for (sentence::const_iterator w=s.begin(); w!=s.end(); w++) sent.push_back(*w);

  int targetIndex=0;
  for (vector<word>::const_iterator w=sent.begin(); w!=sent.end(); w++,targetIndex++) {

    // Generate new example
    set<string> exmp;
    
    for (pRel=rules.begin(); pRel!=rules.end(); pRel++) {

      bool wordMatched = (pRel->Target()!=NULL) && (w->get_form()==string(pRel->Target()));

      if (pRel->TargetIndex()==targetIndex ||  pRel->TargetIndex()==TARGET_ALL || wordMatched) {
        features = pRel->Extract(sent,targetIndex);    // Generate features for example
      }

      // if build=true, we are building the lexicon, all features are added
      //  but if build==false we are encoding new sentences given a lexicon
      //  so we consult the lexicon before adding the feature
      for (pFeat = features.begin(); pFeat != features.end(); pFeat++) 
	if (build || lexicon.find(*pFeat)!=lexicon.end())
	  exmp.insert(*pFeat);
    }

    // add example to result vector, even if empty set (we must keep vector positions 
    // aligned with sentence positions). 
    result.push_back(exmp);
  }
}


////////////////////////////////////////////////////////
/// encode given sentence in feature codes, add
/// result to given vector. 
////////////////////////////////////////////////////////

void fex::encode_int(const sentence &s, std::vector<std::set<int> > &result) {
  vector<RGF>::iterator pRel;
  set<string>::iterator pFeat;
  set<string> features; 
  map<string,int>::const_iterator p;

  if (lexicon_usable == false) {
    cerr<<"ERROR - Integer encoding is not possible using a temporary feature lexicon."<<endl;
    exit(1);
  }

  // copy sentence (list<word>) to vector<word>, needed by FEX classes
  vector<word> sent;
  sent.reserve(s.size());
  for (sentence::const_iterator w=s.begin(); w!=s.end(); w++) sent.push_back(*w);

  int targetIndex=0;
  for (vector<word>::const_iterator w=sent.begin(); w!=sent.end(); w++,targetIndex++) {

    // Create new example for current word
    set<int> exmp;
    
    for (pRel=rules.begin(); pRel!=rules.end(); pRel++) {

      bool wordMatched = (pRel->Target()!=NULL) && (w->get_form()==string(pRel->Target()));

      if (pRel->TargetIndex()==targetIndex ||  pRel->TargetIndex()==TARGET_ALL || wordMatched) {
	// Obtain features for the example
        features = pRel->Extract(sent,targetIndex);   
      }

      // search features in lexicon and translate them if found
      for (pFeat = features.begin(); pFeat != features.end(); pFeat++) {
	p = lexicon.find(*pFeat);
	if (p!=lexicon.end()) exmp.insert(p->second);
      }
    }

    // add example to result vector, even if empty set (we must keep vector positions 
    // aligned with sentence positions). 
    result.push_back(exmp);
  }
}


////////////////////////////////////////////////////////
/// encode given sentence in feature names, return
/// result as vector. Useful for Perl/java APIs
////////////////////////////////////////////////////////

vector<list<string> > fex::encode_name(const sentence &sent, bool build) {

vector<set<string> >::const_iterator p;
set<string>::const_iterator q;

  // encode sentence in temporary vector
  vector<set<string> > res;
  encode_name(sent,res,build);

  vector<list<string> > result;
  for (p=res.begin(); p!=res.end(); p++) {
    // convert set<string> to list<string>, more suitable for Java API
    list<string> a;
    for (q=p->begin(); q!=p->end(); q++) a.push_back(*q);
    // add list to final vector
    result.push_back(a);
  }

  // return vector<list<string> >
  return result;
}



////////////////////////////////////////////////////////
/// encode given sentence in feature codes, return
/// result as vector. Useful for Perl/java APIs
////////////////////////////////////////////////////////

vector<list<int> > fex::encode_int(const sentence &sent) {
vector<set<int> >::const_iterator p;
set<int>::const_iterator q;

  // encode sentence in temporary vector
  vector<set<int> > res;
  encode_int(sent,res);

  vector<list<int> > result;
  for (p=res.begin(); p!=res.end(); p++) {
    // convert set<int> to list<int>, more suitable for Java API
    list<int> a;
    for (q=p->begin(); q!=p->end(); q++) a.push_back(*q);
    // add list to final vector
    result.push_back(a);
  }

  // return vector<list<int> >
  return result;
}

////////////////////////////////////////////////////////
/// Clear all features in the lexicon (either temporary or not)
////////////////////////////////////////////////////////

void fex::clear_lexicon() {
  lexicon.clear();
  lexicon_usable=false;
  sum=0;
}

////////////////////////////////////////////////////////
/// Add a feature name to the temporary lexicon
////////////////////////////////////////////////////////

void fex::add_lexicon(const std::string &name) {
map<string,int>::iterator p;

  if (lexicon_usable == true) {
    cerr<<"ERROR - Features can only be added to temporary lexicons."<<endl;
    exit(1);
  }

  // if feature exists increase count, else init count to 1.
  p = lexicon.find(name);
  if (p != lexicon.end()) p->second++;
  else lexicon.insert(make_pair(name,1));

  // increase total count;
  sum++;
}


////////////////////////////////////////////////////////
/// Load a finished lexicon, filtered and numbered.
////////////////////////////////////////////////////////

void fex::load_lexicon(const std::string &flex) {
  int num, freq;
  string name;

  lexicon.clear();

  ifstream lex(flex.c_str());
  if (!lex) { 
    cerr<<"Error opening file "<<flex<<endl;
    exit(1);
  }

  while (lex>>num>>name>>freq) lexicon.insert(make_pair(name,num));
  lex.close();

  lexicon_usable = true;
}



////////////////////////////////////////////////////////                                
/// Filter, number and save a temporary lexicon.                                              
////////////////////////////////////////////////////////                        

void fex::save_lexicon(const std::string &flex, double min) const {
map<string,int>::const_iterator i;

  if (lexicon_usable == true) {
    cerr<<"ERROR - Only temporary feature lexicons can be saved"<<endl;
    exit(1);
  }

  // if threshold given in relative frequencies, convert to absolut 
  if (min<1) min = min*sum; 
  // feature counter                   
  int n=1;

  ofstream lex(flex.c_str());
  for (i=lexicon.begin(); i!=lexicon.end(); i++) {
    // number and save only relevant features                         
    if (i->second > min)
       lex << n++ <<" "<< i->first <<" "<< i->second << endl;
  }
  lex.close();

}





