//////////////////////////////////////////////////////////////////
//
//    Fries - Feature Retriever for Intensional Encoding of Sentences
//
//    Copyright (C) 2006   TALP Research Center
//                         Universitat Politecnica de Catalunya
//
//    This file is part of the Fries library
//
//    The Fries library is free software; you can redistribute it
//    and/or modify it under the terms of the GNU General Public
//    License as published by the Free Software Foundation; either
//    version 2 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    General Public License for more details.
//
//    You should have received a copy of the GNU General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 51 Franklin St, 5th Floor, Boston, MA 02110-1301 USA
//
//    contact: Lluis Padro (padro@lsi.upc.es)
//             TALP Research Center
//             despatx Omega.S112 - Campus Nord UPC
//             08034 Barcelona.  SPAIN
//
////////////////////////////////////////////////////////////////

#include "fries/language.h"
#include "fries/util.h"

using namespace std;

////////////////////////////////////////////////////////////////
///   Class analysis stores a possible reading (lemma, PoS, probability, distance)
///   for a word.
////////////////////////////////////////////////////////////////

/// Create empty analysis
analysis::analysis() {};
/// Create a new analysis with provided data.
analysis::analysis(const std::string &l, const std::string &p) {lemma=l; parole=p; prob= -1.0; distance= -1.0;};
/// Assignment
analysis& analysis::operator=(const analysis &a) {
  if(this!=&a) {
    lemma=a.lemma; parole=a.parole; prob=a.prob; distance=a.distance;
    senses=a.senses; retok=a.retok; user=a.user;
  }
  return *this;
}
/// Set lemma for analysis.
void analysis::set_lemma(const std::string &l) {lemma=l;}
/// Set parole PoS tag for analysis
void analysis::set_parole(const std::string &p) {parole=p;}
/// Set probability for analysis
void analysis::set_prob(double p) { prob=p;}
/// Set distance for analysis
void analysis::set_distance(double d) { distance=d;}
/// Set retokenization info for analysis
void analysis::set_retokenizable(const std::list<word> &lw) {retok=lw;}
/// Check whether probability has been set.
bool analysis::has_prob(void) const {return(prob>=0.0);}
/// Check whether distance has been set.
bool analysis::has_distance(void) const {return(distance>=0.0);}
/// Get lemma value for analysis.
string analysis::get_lemma(void) const {return(lemma);}
/// Get probability value for analysis (-1 if not set).
double analysis::get_prob(void) const {return(prob);}
/// Get distance value for analysis (-1 if not set).
double analysis::get_distance(void) const {return(distance);}
/// Find out if the analysis may imply retokenization
bool analysis::is_retokenizable(void) const {return(!retok.empty());}
/// Get retokenization info for analysis.
list<word> analysis::get_retokenizable(void) const {return(retok);}
/// Get parole PoS tag value for analysis.
string analysis::get_parole(void) const {return(parole);}
/// Get short version of parole PoS tag value for analysis.
string analysis::get_short_parole(const std::string &lang) const {
  string r;

  // In english, we use WSJ tags, no short version available. Someday we'll move to parole.
  if (lang=="en") return (parole);

  // In other languages, we use parole. Short version is possible.
  // The appropriate length depends on the language and category.
  if (parole[0]=='F') r=parole;
  else if (parole[0]=='V') r.assign(parole.substr(0,3));
  else r.assign(parole.substr(0,2));
  return(r);
}
/// get analysis sense list
list<pair<string, double> > analysis::get_senses(void) const {return(senses);};
string analysis::get_senses_string(void) const {return(util::pairlist2string(senses,":", "/"));};
/// set analiysis sense list
void analysis::set_senses(const std::list<pair<std::string, double> > &ls) {senses=ls;};
/// Comparison to sort analysis by *decreasing* probability
int analysis::operator<(const analysis &a) const {return(prob>a.prob);}


////////////////////////////////////////////////////////////////
///   Class word stores all info related to a word:
///  form, list of analysis, list of tokens (if multiword).
////////////////////////////////////////////////////////////////

/// Create an empty new word
word::word() {}
/// Create a new word with given form.
word::word(const std::string &f) {
  form=f;
  in_dict=true;  // everything is a known word until dictionary fails to find it.
  locked=false;
}
/// Create a new multiword, including given words.
word::word(const std::string &f, const std::list<word> &a) {
  form=f; multiword=a;
  start=a.front().get_span_start();
  finish=a.back().get_span_finish();
  in_dict=true;
  locked=false;
}
/// Create a new multiword, including given words, and with given analysis list.
word::word(const std::string &f, const std::list<analysis> &la, const std::list<word> &a) {
  word::const_iterator i;
  this->clear(); for (i=la.begin(); i!=la.end(); i++) this->push_back(*i);
  form=f; multiword=a;
  start=a.front().get_span_start();
  finish=a.back().get_span_finish();
  in_dict=true;
  locked=false;
  selected=this->begin();
}
/// Copy constructor.
word::word(const word & w) {
  word::const_iterator i;
  this->clear(); copy_analysis(w);
  form=w.form; multiword=w.multiword;
  start=w.start; finish=w.finish;
  in_dict=w.in_dict;
  locked=w.locked;
  user=w.user;
  alternatives=w.alternatives;
  selected=this->begin();
  for (i=w.begin(); i!=w.selected && i!=w.end(); i++)
      selected++;
};
/// Assignment.
word& word::operator=(const word& w){
  word::const_iterator i;
  if(this!=&w) {
    this->clear(); copy_analysis(w);
    form=w.form; multiword=w.multiword;
    start=w.start; finish=w.finish;
    in_dict=w.in_dict;
    locked=w.locked;
    user=w.user;
    alternatives=w.alternatives;
    selected=this->begin();
    for (i=w.begin(); i!=w.selected && i!=w.end(); i++)
      selected++;
  }
  return *this;
};
/// Set (override) word analysis list with one single analysis
void word::set_analysis(const analysis &a) {
   this->clear(); this->push_back(a); selected=this->begin();
}
/// Set (override) word analysis list.
void word::set_analysis(const std::list<analysis> &a) {
  word::const_iterator i;
  this->clear(); for (i=a.begin(); i!=a.end(); i++) this->push_back(*i);
  selected=this->begin();
}
/// Add one analysis to word analysis list.
void word::add_analysis(const analysis &a) {
  this->push_back(a);
  if (this->size()==1) selected=this->begin();
}
/// Set word form.
void word::set_form(const std::string &f) {form=f;}
/// Set token span.
void word::set_span(unsigned long s, unsigned long e) {start=s; finish=e;}

/// get in_dict
bool word::found_in_dict(void) const {return (in_dict);}
/// set in_dict
void word::set_found_in_dict(bool b) {in_dict=b;}
/// check if there is any retokenizable analysis
bool word::has_retokenizable(void) const {
  word::const_iterator i;
  bool has=false;
  for (i=this->begin(); i!=this->end() && !has; i++) has=i->is_retokenizable();
  return(has);
}

/// add an alternative to the alternatives list
void word::add_alternative(const word &w, double d) {
 alternatives.push_back(make_pair(w,d));
}
/// replace alternatives list with list given
void word::set_alternatives(const std::list<std::pair<word,double> > &lw) {
 alternatives=lw;
}
/// find out if the speller checked alternatives
bool word::has_alternatives() const {return (alternatives.size()>0);}
/// get alternatives list
std::list<std::pair<word,double> > word::get_alternatives() const {return(alternatives);}
/// get alternatives begin iterator
std::list<std::pair<word,double> >::iterator word::alternatives_begin() {return alternatives.begin();}
/// get alternatives end iterator
std::list<std::pair<word,double> >::iterator word::alternatives_end() {return alternatives.end();}

/// mark word as having definitive analysis
void word::lock_analysis() { locked=true; }
/// check if word is marked as having definitive analysis
bool word::is_locked(void) const { return locked; }
/// Get length of analysis list.
int word::get_n_analysis(void) const {return(this->size());}
/// Copy analysis list of given word.
void word::copy_analysis(const word &w) {
  list<analysis>::const_iterator i;
  this->clear();
  for (i=w.begin(); i!=w.end(); i++) this->push_back(*i);
  selected=this->begin();
}
/// get list of analysis (only useful for perl API)
list<analysis> word::get_analysis(void) const {return(*this);};
/// get begin iterator to analysis list.
word::iterator word::analysis_begin(void) {return this->begin();}
word::const_iterator word::analysis_begin(void) const {return this->begin();}
/// get end iterator to analysis list.
word::iterator word::analysis_end(void) {return this->end();}
word::const_iterator word::analysis_end(void) const {return this->end();}
/// put all analysis in the selected half list
void word::select_all_analysis() {selected=this->begin();}
/// empty the half list of selected analysis
void word::unselect_all_analysis() {selected=this->end();}
/// Add the given analysis to the selected half.
void word::select_analysis(word::iterator tag) {
  if (selected==this->end()) {
    // none selected. put new selected at the end and update iterators
    this->splice(this->end(), *this, tag);
    selected=this->end(); selected--;
  }
  else if (selected!=this->begin()) {
    // some selected, just move new to the end
    this->splice(this->end(), *this, tag);
  }
  // otherwise, all are selected, do nothing
}
/// remove the given analysis from selected list, moving it to the left of "selected" mark
void word::unselect_analysis(word::iterator tag) {
  if (selected==tag)   // tag to unselect is first in selected list
    selected++;
  else if (selected!=this->end())  // tag is other but first, and there are some selected
    this->splice(selected, *this, tag);
  // else, is none selected, do nothing
}
/// Get the number of selected analysis
int word::get_n_selected(void) const {return(this->size() - get_n_unselected());}
/// Get the number of unselected analysis
int word::get_n_unselected(void) const {
  int n=0;
  for (list<analysis>::const_iterator i=unselected_begin(); i!=unselected_end(); i++)
    n++;
  return(n);
}
/// Check whether the word is a compound.
bool word::is_multiword(void) const {return(!multiword.empty());}
/// Get number of words in compound.
int word::get_n_words_mw(void) const {return(multiword.size());}
/// Get list of words in compound.
list<word> word::get_words_mw(void) const {return(multiword);}
/// Get word form.
string word::get_form(void) const {return(form);}
/// Get the first selected analysis iterator
word::iterator word::selected_begin(void) {return(selected);}
/// Get the first selected analysis iterator
word::const_iterator word::selected_begin(void) const {return(selected);}
/// Get the end of selected analysis list
word::iterator word::selected_end(void) {return(this->end());}
/// Get the end of selected analysis list
word::const_iterator word::selected_end(void) const {return(this->end());}
/// Get the first unselected analysis iterator
word::iterator word::unselected_begin(void) {return (selected==this->begin() ? this->end() : this->begin());}
/// Get the first unselected analysis iterator
word::const_iterator word::unselected_begin(void) const {return (selected==this->begin() ? this->end() : this->begin());}
/// Get the end of unselected analysis list
word::iterator word::unselected_end(void) {return (selected==this->begin() ? this->end() : selected);}
/// Get the end of unselected analysis list
word::const_iterator word::unselected_end(void) const {return (selected==this->begin() ? this->end() : selected);}
/// Get lemma for the selected analysis in list.
string word::get_lemma(void) const {if (this->get_n_analysis()) return(selected->get_lemma());
                                           else return("");}
/// Get parole PoS tag for the selected analysis in list.
string word::get_parole(void) const {if (this->get_n_analysis()) return(selected->get_parole());
                                           else return("");}
/// Get short version of parole PoS tag value for selected analysis.
string word::get_short_parole(const std::string &lang) const {if (this->get_n_analysis())
                                                                  return(selected->get_short_parole(lang));
                                                                else return("");}
/// get sense list for the selected analysis
list<pair<string,double> > word::get_senses(void) const {return(selected->get_senses());};
string word::get_senses_string(void) const {return(util::pairlist2string(selected->get_senses(),":", "/"));};
/// set sense list for the selected analysis
void word::set_senses(const std::list<pair<std::string,double> > &ls) {selected->set_senses(ls);};

/// Get token span.
unsigned long word::get_span_start(void) const {return(start);}
unsigned long word::get_span_finish(void) const {return(finish);}
/// look for a tag in the analysis list of a word
bool word::find_tag_match(RegEx &re) {
  bool found=false;
  for (word::const_iterator an=this->begin(); an!=this->end() && !found; an++) {
    string tag = an->get_parole();
    found = re.Search(tag.c_str());
  }
  return found;
}


////////////////////////////////////////////////////////////////
///   Class node stores nodes of a parse_tree
///  Each node in the tree is either a label (intermediate node)
///  or a word (leaf node)
////////////////////////////////////////////////////////////////

/// Methods for parse_tree nodes
node::node(const std::string & s) : label(s), w(NULL) {head=false;chunk=false;nodeid="-";}
node::node() : label(""), w(NULL) {head=false;chunk=false;nodeid="-";}
string node::get_node_id() const {return (nodeid);}
void node::set_node_id(const string &id) {nodeid=id;}
string node::get_label() const {return(label);}
word node::get_word() const {return (*w);}
void node::set_label(const std::string &s) {label=s;}
void node::set_word(word &wd)  {w = &wd;}
bool node::is_head() const {return head;}
void node::set_head(const bool h) {head=h;}
bool node::is_chunk() const {return (chunk!=0);}
void node::set_chunk(const int c) {chunk=c;}
int  node::get_chunk_ord() const {return (chunk);}


////////////////////////////////////////////////////////////////
///   Class parse tree is used to store the results of parsing
////////////////////////////////////////////////////////////////

int parse_tree::nsentence=0;
/// Methods for parse_tree
parse_tree::parse_tree() {};
parse_tree::parse_tree(parse_tree::iterator p) :tree<node>(p) {};
parse_tree::parse_tree(const node & n) : tree<node>(n) {};
/// assign id's to nodes and build index
void parse_tree::build_node_index() {
  parse_tree::iterator k;
  int i=0;
  string sid=util::int2string(nsentence);
  node_index.clear();
  for (k=this->begin(); k!=this->end(); ++k, i++) {
    string id=sid+"."+util::int2string(i);
    k->info.set_node_id(id);
    node_index.insert(make_pair(id,k));
  }
  nsentence++;
};
// rebuild index maintaining id's
void parse_tree::rebuild_node_index() {
  parse_tree::iterator k;
  node_index.clear();
  for (k=this->begin(); k!=this->end(); ++k) {
    string id=k->info.get_node_id();
    if (id != "-") node_index.insert(make_pair(id,k));
  }
};
/// get node with given index.
parse_tree::iterator parse_tree::get_node_by_id(string id) const {
  map<std::string,parse_tree::iterator>::const_iterator p;
  p = node_index.find(id);
  if (p!=node_index.end()) return p->second;
  else return NULL;
};


////////////////////////////////////////////////////////////////
/// class denode stores nodes of a dependency tree and
///  parse tree <-> deptree relations
////////////////////////////////////////////////////////////////

/// Methods for dependency tree nodes
depnode::depnode() {};
depnode::depnode(const std::string & s) : node(s) {}
depnode::depnode(const node & n) : node(n) {}
void depnode::set_link(const parse_tree::iterator p) {link=p;}
parse_tree::iterator depnode::get_link(void) { return link;}
tree<node>& depnode::get_link_ref(void) { return (*link);}  ///  (useful for Java API)


////////////////////////////////////////////////////////////////
/// class dep_tree stores a dependency tree
////////////////////////////////////////////////////////////////

/// Constructors for dep_tree
dep_tree::dep_tree() {};
dep_tree::dep_tree(const depnode & n) : tree<depnode>(n) {};


////////////////////////////////////////////////////////////////
///   Class sentence is just a list of words that someone
/// (the splitter) has validated it as a complete sentence.
/// It may include a parse tree.
////////////////////////////////////////////////////////////////

/// Create a new sentence.
sentence::sentence() {};
/// Create a new sentence from list<word>
sentence::sentence(const list<word> &lw) : list<word>(lw) {};
/// Set the parse tree.
void sentence::set_parse_tree(const parse_tree &tr) { pt=tr;};
/// Obtain the parse tree.
parse_tree & sentence::get_parse_tree(void) {return(pt);};
/// Find out whether the sentence is parsed.
bool sentence::is_parsed() const {return !pt.empty();};
/// Set the dependency tree.
void sentence::set_dep_tree(const dep_tree &tr) { dt=tr;};
/// Obtain the parse dependency tree
dep_tree & sentence::get_dep_tree() {return(dt);};
/// Find out whether the sentence is dependency parsed.
bool sentence::is_dep_parsed() const {return !dt.empty();};
/// obtain list of words (useful for perl APIs)
vector<word> sentence::get_words() const {
  vector<word> v;
  for (sentence::const_iterator i=this->begin(); i!=this->end(); i++)
    v.push_back(*i);
  return (v);
};
/// obtain iterators (useful for perl/java APIs)
sentence::iterator sentence::words_begin(void) {return this->begin();};
sentence::const_iterator sentence::words_begin(void) const {return this->begin();};
sentence::iterator sentence::words_end(void) {return this->end();};
sentence::const_iterator sentence::words_end(void) const {return this->end();};


////////////////////////////////////////////////////////////////
///   Class paragraph is just a list of sentences that someone
///  has validated it as a paragraph.
////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////
///   Class document is a list of paragraphs. It may have additional
///  information (such as title)
////////////////////////////////////////////////////////////////

/// Constructor
document::document() {}

/// Add node1 to the group group1 of coreferents
void document::add_positive(string node1, int group1) {
  group2node.insert( make_pair(group1,node1) );
  node2group[node1] = group1;
}

/// Add second node to coreference group of the first (or in a new group if first didn't have any)
void document::add_positive(string node1, string node2) {
  // find out group of node1, creating a new group if it didn't have any.
  int g1 = get_coref_group(node1);
  if (g1 == -1) {
    g1 = ++last_group;
    group2node.insert( make_pair(g1,node1) );
    node2group[node1] = g1;
  }
  // add second node to group g1
  group2node.insert( make_pair(g1,node2) );
  node2group[node2] = g1;
}

/// Gets the id of the coreference group of the node
int document::get_coref_group(string node1) const {
  std::map<string,int>::const_iterator it = node2group.find(node1);
  if (it == node2group.end()) return -1;
  else return it->second;
}

/// Gets all the nodes in a coreference group id
std::list<string> document::get_coref_nodes(int id) const {
  std::list<string> ret;
  std::multimap<int,string>::const_iterator it;
  std::pair<std::multimap<int,string>::const_iterator,std::multimap<int,string>::const_iterator> par;
  par = group2node.equal_range(id);
  for (it=par.first; it!=par.second; ++it) ret.push_back(it->second);
  return(ret);
}

/// Returns whether two nodes are in the same coreference group
bool document::is_coref(string node1, string node2) const {
  int g1, g2;
  g1 = get_coref_group(node1);
  g2 = get_coref_group(node2);
  return ( g1!= -1 && g1==g2 );
}
